"""
Module to declare the user-defined available datahandlers.

All datahandlers must be imported here and added to the `available_datahandlers` dictionary containing {"datahandler_type": Datahandler} to be available in the catalog.
"""
from canonada.catalog import available_datahandlers

available_datahandlers.update({})