"""
Manage the data catalog for the canonada.
"""

from ._core import get as get
from ._core import ls as ls
from ._core import params as params
from ._core import credentials as credentials
from ._datahandlers import Datahandler as Datahandler
from ._datahandlers import available_datahandlers as available_datahandlers
