"""
Canonada

A data science framework that helps you build production-ready streaming pipelines for data processing in Python
"""

from . import catalog as catalog
from . import pipeline as pipeline
from . import system as system